@@ -35,7 +35,9 @@
 #endif
 
 /* SunOS 4.1.4 patch from Tom Lipkis <tal@pss.com> */
-#if defined(__sun__) && defined(__sparc__) && !defined(__svr4__)
+#if (defined(__sun__) && defined(__sparc__) && !defined(__svr4__)) /* SunOS */
+     || (defined(_IRIX_))  /* IRIX */
+     || (defined(__osf__)) /* Digital UNIX */
 # ifndef EX__MAX
 # define EX__MAX 77
 extern char *optarg;
@@ -208,7 +210,7 @@ int read_message(int in, int out, int max_size)
   int header_read=0;
   char buf[8192];
   char is_spam[5];
-  int score,threshold;
+  float score,threshold;
   float version;
   int response=EX_OK;
   char* out_buf;
@@ -268,12 +270,12 @@ int read_message(int in, int out, int max_size)
 	  if(CHECK_ONLY)
 	  {
 	    /* Ok, found a header line, it better be "Spam: x; y / x" */
-	    if(3 != sscanf(buf,"Spam: %5s ; %d / %d",is_spam,&score,&threshold))
+	    if(3 != sscanf(buf,"Spam: %5s ; %f / %f",is_spam,&score,&threshold))
 	    {
 	      response = EX_PROTOCOL; break;
 	    }
 
-	    printf("%d/%d\n",score,threshold);
+	    printf("%.1f/%.1f\n",score,threshold);
 
 	    if(!strcasecmp("true",is_spam)) /* If message is indeed spam */
 	    {
@@ -479,6 +481,7 @@ int process_message(const char *hostname, int port, char *username, int max_size
 
     if(CHECK_ONLY && ESC_PASSTHROUGHRAW == exstatus)
     {
+	printf("0/0\n");
 	exstatus = EX_OK;
     }
 
